<?php

   /* Open Sound Control Client/Server Utility
    * September 5th, 2003, Version 1.0
    * Please see included osc_hub.README.txt file for more information.
    *
    * Copyright 2003, Michal Migurski <mike@teczno.com> <http://mike.teczno.com>
    * 
    * This library is free software; you can redistribute it and/or modify it under
    * the terms of the GNU Lesser General Public License as published by the Free
    * Software Foundation; either version 2.1 of the License, or (at your option) any
    * later version.
    * 
    * This library is distributed in the hope that it will be useful, but WITHOUT ANY
    * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
    * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    * 
    * You should have received a copy of the GNU Lesser General Public License along
    * with this library; if not, write to the Free Software Foundation, Inc., 59
    * Temple Place, Suite 330, Boston, MA 02111-1307 USA
    */

    class osc_hub
    {
        var $recv_addr;
        var $recv_port;
        var $send_addr;
        var $send_port;

        var $recv_sock;
        var $send_sock;
        
       /** osc_hub
        * constructor function for an OSC
        * respond-and-reply handler.
        *
        * @param    string  recv    address and port to receive on,
        *                           must be in address:port form.
        *
        * @param    string  send    address and port to send to,
        *                           must be in address:port form.
        */
        function osc_hub($recv, $send)
        {
            // Take note of the configuration for this machine.
            // osc_hub::_byte_order_test();
            
            if(preg_match('/[\w\.]+:\d+/', $recv)) list($this->recv_addr, $this->recv_port) = explode(':', $recv);
            if(preg_match('/[\w\.]+:\d+/', $send)) list($this->send_addr, $this->send_port) = explode(':', $send);
        }
        
       /** listen
        * listen for incoming OSC messages
        *
        * @param    string  callback    callback function, which must take 2 arguments:
        *                               array: bundle received from remote sender
        *                               object: a reference to an osc_hub
        *                               if result of callback function evaluates to
        *                               false, listen() will stop.
        */
        function listen($callback)
        {
            if($this->recv_addr) {
                if(($this->recv_sock = socket_create(AF_INET, SOCK_DGRAM, 0)) < 0)
                    trigger_error(sprintf("socket_create() failed, reason: %s\n" . socket_strerror($this->recv_sock)), E_USER_ERROR);
                
                if(($ret = socket_bind($this->recv_sock, $this->recv_addr, $this->recv_port)) < 0)
                    trigger_error(sprintf("socket_bind() failed, reason: %s\n" . socket_strerror($ret)), E_USER_ERROR);
                
            } else {
                trigger_error("osc_hub->listen() failed: you need to specify a source address.\n", E_USER_ERROR);
                
            }
    
            while(true) {
                if(false === ($buffer = socket_read($this->recv_sock, 2048, PHP_BINARY_READ))) {
                    trigger_error(sprintf("socket_read() failed, reason: %s\n" . socket_strerror($ret)), E_USER_ERROR);
                    break;
                }
                
                if($GLOBALS['_VERBOSE'])
                    echo osc_hub::_buf2str($buffer);

                if(!call_user_func($callback, osc_hub::_buf2arr($buffer, $GLOBALS['_VERBOSE']), $this))
                    break;
            }
        
            socket_close($this->send_sock);
            $this->send_sock = false;
        }
        
       /** send
        * send outgoing OSC messages
        *
        * @param    mixed   bundle  bundle of OSC data, always
        *                           an array of arrays of atoms.
        *                           a string that is acceptable to
        *                           _str2arr() will also work here.
        *
        * @return   bool    true or false, depending on success of socket_sendto()
        */
        function send($bundle)
        {
            if(is_string($bundle))
                $bundle = osc_hub::_str2arr($bundle);
                
            assert('is_array($bundle)');
            $buffer = osc_hub::_arr2buf($bundle, $GLOBALS['_VERBOSE']);

            if($GLOBALS['_VERBOSE'])
                echo osc_hub::_buf2str($buffer);

            if(!$this->send_sock && $this->send_addr)
                if(($this->send_sock = socket_create(AF_INET, SOCK_DGRAM, SOL_UDP)) < 0)
                    trigger_error(sprintf("socket_create() failed, reason: %s\n" . socket_strerror($this->send_sock)), E_USER_ERROR);

            if(false === socket_sendto($this->send_sock, $buffer, strlen($buffer), 0, $this->send_addr, $this->send_port)) {
                trigger_error("osc_hub->send() failed: unknown reason.\n", E_USER_ERROR);
                return false;
            }
            
         // give it a rest for a few microseconds
            usleep(100);
            
            return true;
        }

       /** _str2arr
        * converts string representation to bundle
        *
        * @param    string  text_bundle     text representation of a bundle,
        *                                   one message per line, delimited by spaces
        *
        * @return   array   array representation of bundle
        */
        function _str2arr($text_bundle)
        {
            $bundle = array();
    
            foreach(explode("\n", $text_bundle) as $m => $message)
                if(!empty($message))
                    foreach(split(" ", $message) as $a => $atom)
                        if(($atom == '0') || !empty($atom))
                            if(is_numeric($atom)) {
                                if(strpos($atom, '.')) {
                                    $bundle[$m][$a] = floatval($atom);
                                    
                                } else {
                                    $bundle[$m][$a] = intval($atom);
                                    
                                }
    
                            } else {
                                $bundle[$m][$a] = $atom;
    
                            }
                    
            if($GLOBALS['_VERBOSE']) print_r($bundle);
            return $bundle;
        }
    
       /** _buf2str
        * convert buffer to string representation
        *
        * @param    string  buffer  binary data string
        *
        * @return   string  multi-line ascii string showing
        *                   hex values of input buffer
        */
        function _buf2str($buffer)
        {
            $output = '';
            
            $hexbuf = bin2hex($buffer);
            $bufpos = 0;
            $hexstr = '';
            $ascstr = '';
            $format = " %-49s |%-16s|\n";
        
            do {
                if($bufpos) {
                    if(0 == ($bufpos % 16)) {
                        $output .= sprintf($format, $hexstr, $ascstr);
                        $hexstr = '';
                        $ascstr = '';
        
                    } elseif(0 == ($bufpos % 8)) {
                        $hexstr .= ' ';
        
                    }
                }
                $bufpos++;
        
                $hexbyte = substr($hexbuf, 0, 2);
                $hexstr .= $hexbyte.' ';
        
                if((hexdec($hexbyte) >= 32) && (hexdec($hexbyte) <= 126)) {
                    $ascstr .= chr(hexdec($hexbyte));
                } else {
                    $ascstr .= '.';
                }
        
            } while($hexbuf = substr($hexbuf, 2));
        
            $output .= sprintf("$format\n", $hexstr, $ascstr);
            return $output;
        }
        
       /** _buf2arr
        * convert buffer to bundle array representation
        *
        * @param    string  buffer      binary data string
        *
        * @param    bool    verbose     echo chatty diagnostic info?
        *
        * @return   array   array of arrays of atoms
        */
        function _buf2arr($buffer, $verbose=false)
        {
            ob_start();
            $bundle = array();
        
            for($bufpos = 0; $bufpos < strlen($buffer); ) {
                if(substr($buffer, $bufpos, 7) == '#bundle') {
                   /* found the beginning of a bundle, huzzah.
                    * skip the time tag and move on to its contents on the next iteration
                    */
                    echo "bundle!\n";
                    $bufpos += 16;
                    continue;
        
                } else {
                   /* we should be at the start of a message here,
                    * and the first 4 bytes should give its length.
                    */
                    $len = reset(unpack('i', osc_hub::_byte_order_fix(substr($buffer, $bufpos, 4))));
                    $bufpos += 4;
        
                    $subbuf = substr($buffer, $bufpos, $len);
                    $format = false;
                    $tokens = array();
                    $curtoken = 0;
                 // printf("bundle of length %d: %s\n", $len, bin2hex($subbuf));
                    
                   /* iterate over each 4-byte chunk of the current message,
                    * appending to a token array as nulls are encountered.
                    */
                    for($subbufpos = 0; $subbufpos < strlen($subbuf); $subbufpos += 4) {
                       /* append the current sub-buffer chunk to the last
                        * element of the tokens array.
                        */
                        $bufchunk = substr($subbuf, $subbufpos, 4);
                        $tokens[$curtoken] = empty($tokens[$curtoken])
                                               ? $bufchunk
                                               : $tokens[$curtoken] . $bufchunk;
        
                       /* check whether we are at a token boundary:
                        * end of the sub-buffer, or just at a null character
                        */
                        if((($subbufpos + 4) >= strlen($subbuf)) || strpos($bufchunk, "\0") || (strpos($bufchunk, "\0") === 0)) {
                         // printf("finished token %d: %s\n", $curtoken, bin2hex($tokens[$curtoken]));
        
                           /* check whether the current token is the format string
                            */
                            if(!$format && (strpos($tokens[$curtoken], ',') === 0)) {
                               /* found the format string:
                                * move its data to the format array,
                                * remove it from the tokens array,
                                * ensure that array indexes match up.
                                */
                                $format = preg_split('//', substr(trim($tokens[$curtoken]), 1), -1, PREG_SPLIT_NO_EMPTY);
                                array_unshift($format, 's');
                                array_pop($tokens);
                                $curtoken--;
                             // printf("found format string.\n");
                            
                            }
                            
                            if(isset($format[$curtoken]))
                               /* if a format string has been found,
                                * unpack the corresponding token's data into
                                * something more usable.
                                */
                                switch($format[$curtoken]) {
                                    case 's':
                                        $tokens[$curtoken] = trim($tokens[$curtoken]);
                                        break;
            
                                    case 'i':
                                        $tokens[$curtoken] = end(unpack('i', osc_hub::_byte_order_fix($tokens[$curtoken])));
                                        break;
            
                                    case 'f':
                                        $tokens[$curtoken] = end(unpack('f', osc_hub::_byte_order_fix($tokens[$curtoken])));
                                        break;
                                        
                                    default:
                                        $tokens[$curtoken] = '-undefined-';
                                }
        
                            $curtoken++;
                        
                        }
                    
                    }
                    
                   /* bop it all out.
                    */
                    echo join(', ', array_map(create_function('$t, $f',
                                                              'return sprintf("(%s) %s",
                                                                              (($f == "s") ? "string" : (($f == "i") ? "int" : (($f == "f") ? "float" : "-undefined-"))),
                                                                              $t);'),
                                              $tokens,
                                              $format))
                          . "\n";
        
                    $bufpos += $len;
                    $bundle[] = $tokens;
                    continue;
        
                }
        
               /* this line should never be reached, but it's
                * here to help discourage infinite looping.
                */
                $bufpos++;
            }
            
            if($verbose) { ob_end_flush(); } else { ob_end_clean(); }
            return $bundle;
        }
        
       /** _arr2buf
        * convert buffer to bundle array representation
        *
        * @param    array   bundle      array of arrays of atoms
        *
        * @param    bool    verbose     echo chatty diagnostic info?
        *
        * @return   string  binary data string
        */
        function _arr2buf($bundle, $verbose=false)
        {
            ob_start();
            $buffer = "#bundle\0\0\0\0\0\0\0\0\x01";
        
           /* break up incoming bundle into separate messages
            */
            foreach($bundle as $message) {
                if(!is_array($message)) {
                    trigger_error("osc_hub::_arr2buf() error: ${message} needs to be an array.\n", E_USER_ERROR);
                    continue;
                                
                }
    
               /* pull address off of current message
                */
                $address = array_shift($message);
    
               /* pad address to 4-byte lengths
                */
                switch(strlen($address) % 4) {
                    case 0: $address .= "\0";
                    case 1: $address .= "\0";
                    case 2: $address .= "\0";
                    case 3: $address .= "\0";
                }
    
                printf("\naddress %s %s\n", $address, osc_hub::_prettybin($address));
            
               /* break up remainder of message into separate arguments
                */
                $format = ',';
                $args = '';
                foreach($message as $arg) {
                    if(is_array($arg)) {
                        trigger_error("osc_hub::_arr2buf() error: current argument needs to be a string, float, or integer.\n", E_USER_ERROR);
                        continue;
                                    
                    }
    
                   /* convert arguments to bytecode representation:
                    * - translate floats, ints to 32-bit network-order
                    * - pad strings to 4-byte lengths
                    */
                    if(is_float($arg)) {
                        $format .= 'f';
                        printf("    float %f %s\n", $arg, osc_hub::_prettybin(osc_hub::_byte_order_fix(pack('f', $arg))));
                        $arg = osc_hub::_byte_order_fix(pack('f', $arg));
                    
                    } else if(is_int($arg)) {
                        $format .= 'i';
                        printf("    int %d %s\n", $arg, osc_hub::_prettybin(osc_hub::_byte_order_fix(pack('i', $arg))));
                        $arg = osc_hub::_byte_order_fix(pack('i', $arg));
                    
                    } else {
                        $format .= 's';
    
                        switch(strlen($arg) % 4) {
                            case 0: $arg .= "\0";
                            case 1: $arg .= "\0";
                            case 2: $arg .= "\0";
                            case 3: $arg .= "\0";
                        }
    
                        printf("    string %s %s\n", $arg, osc_hub::_prettybin($arg));
                               
                    }
                    
                   /* compose arguments into complete binary string
                    */
                    $args .= $arg;
                    
                }
        
               /* pad format to 4-byte lengths
                */
                switch(strlen($format) % 4) {
                    case 0: $format .= "\0";
                    case 1: $format .= "\0";
                    case 2: $format .= "\0";
                    case 3: $format .= "\0";
                }
    
                printf("format %s %s\n", $format, osc_hub::_prettybin($format));
        
                $subbuf = $address . $format . $args;
                $subbuf = osc_hub::_byte_order_fix(pack('i', strlen($subbuf))) . $subbuf;
        
                printf("subbuf %s\n", osc_hub::_prettybin($subbuf));
                       
               /* compose messages into complete bundle
                */
                $buffer .= $subbuf;
            }
            
            echo "\n";
            
            if($verbose) { ob_end_flush(); } else { ob_end_clean(); }
            return $buffer;
        }
        
       /** _prettybin
        * binary string into nice hex representation
        * suitable for framing.
        *
        * @param    string  bin     binary data string
        *
        * @return   string  formatted binary data string
        */
        function _prettybin($bin)
        {
            return sprintf('(%d, %s)', strlen($bin), bin2hex($bin));
        }
    
       /** _byte_order_fix
        * convert local to network (big endian) byte order
        * This function is adapted from php-osc (http://www.a2hd.com/software/)
        *
        * @param    string      loc_bin     binary string in local byte order
        *
        * @return   string      binary string in network byte order
        */
        function _byte_order_fix($loc_bin)
        {
            $loc_int = pack("L", 1);  // local byte order
            $net_int = pack("N", 1);  // network byte order
            
            if($loc_int[0] != $net_int[0]) {
                $net_bin = "";
                for($i = 0; $i < strlen($loc_bin); $i++)
                    $net_bin .= $loc_bin[(strlen($loc_bin)-1)-$i];
    
                return $net_bin;
            }
    
            return $loc_bin;
        }
    
       /** _byte_order_test
        * sets $GLOBALS['_LITTLE_ENDIAN_BYTE_ORDER'] depending
        * on local byte order
        */
        /*
        function _byte_order_test()
        {
            $loc_int = pack("L", 1);  // local byte order
            $net_int = pack("N", 1);  // network byte order
            
            $GLOBALS['_LITTLE_ENDIAN_BYTE_ORDER'] = 
                ($loc_int[0] == $net_int[0])
                  ? false
                  : true;
        }
        */
    
    }

?>